@@ -292,8 +292,8 @@ class StructuredWave(AcousticField):
                 f"!number format := short float\n"
                 f"!number of bytes per pixel := 4\n"
                 f"scaling factor (mm/pixel) [1] := {self.params['dx'] * 1000}\n"
-                f"scaling factor (mm/pixel) [2] := {self.params['dx'] * 1000}\n"
-                f"scaling factor (s/pixel) [3] := {1 / self.params['f_AQ']}\n"
+                f"scaling factor (mm/pixel) [2] := {self.params['dz'] * 1000}\n"
+                f"scaling factor (s/pixel) [3] := {1 / self.params['f_saving']}\n"
                 f"first pixel offset (mm) [1] := {self.params['Xrange'][0] * 1e3}\n"
                 f"first pixel offset (mm) [2] := {self.params['Zrange'][0] * 1e3}\n"
                 f"first pixel offset (s) [3] := 0\n"
