@@ -700,29 +700,46 @@ class AlgebraicRecon(Recon):
         if withTumor:
             if self.experiment.AOsignal_withTumor is None:
                 raise ValueError("AO signal with tumor is not available. Please generate AO signal with tumor the experiment first in the experiment object.")
-            else:
-                y = self.experiment.AOsignal_withTumor
         else:
             if self.experiment.AOsignal_withoutTumor is None:
                 raise ValueError("AO signal without tumor is not available. Please generate AO signal without tumor the experiment first in the experiment object.")
-            else:
-                y = self.experiment.AOsignal_withoutTumor
 
         if self.optimizer.value == OptimizerType.MLEM.value:
-            self.reconPhantom, self.indices = MLEM(SMatrix=self.SMatrix, 
-                                                    y=y,
-                                                    numIterations=self.numIterations,
-                                                    isSavingEachIteration=self.isSavingEachIteration,
-                                                    withTumor=withTumor,
-                                                    use_multi_gpu= self.isMultiGPU,
-                                                    use_numba= self.isMultiCPU,
-                                                    max_saves=self.maxSaves
-                                                    )
+            if withTumor:
+                self.reconPhantom, self.indices = MLEM(SMatrix=self.SMatrix, 
+                                                        y=self.experiment.AOsignal_withTumor,
+                                                        numIterations=self.numIterations,
+                                                        isSavingEachIteration=self.isSavingEachIteration,
+                                                        withTumor=withTumor,
+                                                        use_multi_gpu= self.isMultiGPU,
+                                                        use_numba= self.isMultiCPU,
+                                                        max_saves=self.maxSaves
+                                                        )
+            else:
+                self.reconLaser, self.indices = MLEM(SMatrix=self.SMatrix, 
+                                                        y=self.experiment.AOsignal_withoutTumor,
+                                                        numIterations=self.numIterations,
+                                                        isSavingEachIteration=self.isSavingEachIteration,
+                                                        withTumor=withTumor,
+                                                        use_multi_gpu= self.isMultiGPU,
+                                                        use_numba= self.isMultiCPU,
+                                                        max_saves=self.maxSaves
+                                                        )
         elif self.optimizer.value == OptimizerType.LS.value:
             if self.alpha is None:
                 raise ValueError("Alpha (regularization parameter) must be set for LS reconstruction.")
-            self.reconPhantom, self.indices = LS(SMatrix=self.SMatrix,
-                                                y=y, 
+            if withTumor:
+                self.reconPhantom, self.indices = LS(SMatrix=self.SMatrix,
+                                                y=self.experiment.AOsignal_withTumor, 
+                                                numIterations=self.numIterations, 
+                                                isSavingEachIteration=self.isSavingEachIteration, 
+                                                withTumor=withTumor,
+                                                alpha=self.alpha,
+                                                max_saves=self.maxSaves,
+                                                )
+            else:
+                self.reconLaser, self.indices = LS(SMatrix=self.SMatrix,
+                                                y=self.experiment.AOsignal_withoutTumor, 
                                                 numIterations=self.numIterations, 
                                                 isSavingEachIteration=self.isSavingEachIteration, 
                                                 withTumor=withTumor,
