@@ -148,12 +148,6 @@ class PrimalDualRecon(AlgebraicRecon):
 
 
     def _convexReconPython(self, withTumor):
-        if withTumor:
-            y=self.experiment.AOsignal_withTumor
-
-        else:
-            y=self.experiment.AOsignal_withoutTumor
-
         if self.optimizer == OptimizerType.CP_TV:
             if withTumor:
                 self.reconPhantom, self.indices = CP_TV(
